import pyglet

import cocos
from cocos.director import director
from cocos.menu import *

from gamelib import character, DummyPlayer

class InventoryLayer(cocos.layer.Layer):
    is_event_handler = True

    def __init__(self, party, character=None):
        super(InventoryLayer, self).__init__()
        self.party = party
        if character is None:
            character = party.characters[0]
        self.character = character

    def add_heading(self, text, x, y):
        return self.add_label(text, x, y, True)

    def add_label(self, text, x, y, bold=False, color=(0, 0, 0, 255)):
        l = cocos.text.Label(text, (x, y), bold=bold, color=color,
            anchor_y='top')
        self.add(l)
        return l, y - l.element.content_height

    def on_enter(self):
        super(InventoryLayer, self).on_enter()

        self.batch = pyglet.graphics.Batch()

        w, h = director.get_window_size()

        self.add_heading('Backpack', w//3, h)

        self.set_character()
        if self.parent.party_mode:
            self.set_party()

        # add character buttons
        l, y = self.add_heading('Select Character:', w//3, h // 2)
        self.character_names = []
        for character in self.party.characters:
            l, y = self.add_label(character.name, w//3, y)
            l._character = character
            self.character_names.append(l)

        l, y = self.add_heading('Hints:', w//3, y-32)
        l, y = self.add_label("Click on a character's name above to select.", w//3, y)
        l, y = self.add_label('Drag and drop items to move them around.', w//3, y)
        if self.parent.party_mode:
            # party management
            l, y = self.add_label("Click on a character's skill to add experience.", w//3, y)
        l, y = self.add_label('Press <esc> when you are finished.', w//3, y)

        try:
            self.music_player = pyglet.media.Player()
            self.shopping_music = pyglet.resource.media('music_town_mp.ogg')
        except pyglet.media.riff.WAVEFormatException:
            self.music_player = DummyPlayer()
            self.shopping_music = None

        self.music_player.queue(self.shopping_music)
        self.music_player.eos_action = self.music_player.EOS_LOOP
        self.music_player.play()

    def set_character(self):
        w, h = director.get_window_size()

        b = self.character.body

        self.character_slots = []
        self.character_labels = []
        self.skills_labels = []

        def add_head(text, y):
            l, y = self.add_heading(text, 0, y)
            self.character_labels.append(l)
            return y

        y = add_head(self.character.name, h)

        for slot, slabel in zip(b.slots, b.slot_labels):
            item = b.get(slot)
            if item is None:
                label = ''
            else:
                label = item.label
            l, y = self.add_label(slabel + ': ' + label, 0, y)
            l._body_slot = slot
            l._item = item
            self.character_slots.append(l)

        # ATTRIBUTES
        y = add_head('Character Attributes:', y - 16)
        for stat in 'strength strength dexterity constitution intelligence wisdom charisma'.split():
            l, y = self.add_label('%s: %s'%(stat.title(), getattr(self.character,
                stat)), 0, y)
            l._body_slot = slot
            l._item = item
            self.character_slots.append(l)

        # EXPERIENCE
        y = add_head('Experience:', y - 16)
        self.cur_xp, y = self.add_label(str(self.character.experience), 0, y)

        # SKILLS
        y = add_head('Character Skills:', y - 16)
        for skill,value in self.character.skills.items():
            l, y = self.add_label('%s: %s'%(character.skills[skill]['label'], value), 0, y)
            self.skills_labels.append((l, skill))
        if self.parent.party_mode:
            if self.character.experience >= 5:
                l, y = self.add_label('Buy New Skill (5 XP required)', 0, y)
                self.skills_labels.append((l, 'buy skill'))
            else:
                l, y = self.add_label('Buy New Skill (5 XP required)', 0, y,
                    color=(100, 100, 100, 255))
                self.character_labels.append(l)

        # ABILITIES
        l, y = self.add_heading('Character Abilities:', 0, y - 16)
        self.character_labels.append(l)
        if self.parent.party_mode:
            if self.character.experience >= 5:
                l, y = self.add_label('Buy New Ability (5 XP required)', 0, y)
                self.skills_labels.append((l, 'buy ability'))
            else:
                l, y = self.add_label('Buy New Ability (5 XP required)', 0, y,
                    color=(100, 100, 100, 255))
                self.character_labels.append(l)

        # pack contents
        x = w//3
        y = h

        y -= self.character_labels[0].element.content_height

        self.pack_top = y
        self.pack_items = []
        for item in self.character.carried:
            l, y = self.add_label(item.label, x, y)
            l._item = item
            self.pack_items.append(l)

    party_items = None
    def set_party(self):
        w, h = director.get_window_size()

        items = self.party.items

        x = 2 * w//3
        l, y = self.add_heading('Party Items', x, h)

        l, y = self.add_label('%d coins'%self.party.coins, x, y)

        # sort the items and get counts of duplicates
        items.sort()
        last = (None, None)
        l = []
        for item in items:
            if item == last[0]:
                last[1] += 1
            else:
                last = [item, 1]
                l.append(last)

        self.party_items = []
        for item, count in l:
            l, y = self.add_label('%d x %s'%(count, item.label), x, y)
            l._item = item
            l._count = count
            self.party_items.append(l)

        self.party_top = y


    def find_hit(self, x, y):
        for label in self.character_slots:
            if x < label.x: continue
            if y > label.y: continue
            if x > label.x + label.element.content_width: continue
            if y < label.y - label.element.content_height: continue
            if not label._item: return None
            slot = label._body_slot
            item = self.character.body.get(slot)
            return (label, slot)

        for label in self.pack_items:
            if x < label.x: continue
            if y > label.y: continue
            if x > label.x + label.element.content_width: continue
            if y < label.y - label.element.content_height: continue
            return (label, None)

        if not self.parent.party_mode:
            return None

        for label in self.party_items:
            if x < label.x: continue
            if y > label.y: continue
            if x > label.x + label.element.content_width: continue
            if y < label.y - label.element.content_height: continue
            return (label, None)

        return None

    __over = None
    def on_mouse_motion(self, x, y, dx, dy):
        hit = self.find_hit(x, y)
        if not hit:
            if self.__over is not None:
                self.__over.element.color = (0, 0, 0, 255)
                self.__over = None
            return True

        label, slot = hit

        if label is not self.__over:
            if self.__over is not None:
                self.__over.element.color = (0, 0, 0, 255)
            self.__over = label
            label.element.color = (100, 100, 0, 255)

        return True

    __dragging = None
    __drag_start = None
    drag_label = None
    def on_mouse_drag(self, x, y, dx, dy, buttons, modifiers):
        if not self.__dragging:
            # do the mouse-over highlighting
            self.on_mouse_motion(x, y, dx, dy)
            return True

        # XXX highlight the section of screen active

        if self.drag_label:
            self.drag_label.position = (x, y)
            return True

        dx, dy = self.__drag_start
        if abs(x - dx) + abs(y - dy) < 2:
            return True

        item, slot = self.__dragging
        self.drag_label, y = self.add_label(item._item.label, x, y)
        return True

    def on_mouse_press(self, x, y, buttons, modifiers):
        if not buttons & pyglet.window.mouse.LEFT:
            return False

        for l in self.character_names:
            if x < l.x: continue
            if y > l.y: continue
            if x > l.x + l.element.content_width: continue
            if y < l.y - l.element.content_height: continue
            self.character = l._character
            for l in self.character_labels + self.character_slots + self.pack_items:
                self.remove(l)
            for l, skill in self.skills_labels:
                self.remove(l)
            self.remove(self.cur_xp)
            self.set_character()
            return True

        # no experience assignment in encounter mode
        if self.parent.party_mode:
            for l, skill in self.skills_labels:
                if x < l.x: continue
                if y > l.y: continue
                if x > l.x + l.element.content_width: continue
                if y < l.y - l.element.content_height: continue
                if skill == 'buy skill':
                    director.push(cocos.scene.Scene(BuySkillMenu('Buy Skill',
                        self.character)))
                    return True
                elif skill == 'buy ability':
                    print 'BUY ABILITY (NOT IMPLEMENTED)'
                    return True
                elif (self.character.experience > 0 and
                        self.character.skills[skill] < 10):
                    self.character.experience -= 1
                    self.cur_xp.element.text = str(self.character.experience)
                    self.character.skills[skill] += 1
                    l.element.text = '%s: %s'%(character.skills[skill]['label'],
                        self.character.skills[skill])
                    return True


        hit = self.find_hit(x, y)
        if not hit or not hit[0]._item:
            return False

        self.__dragging = hit
        self.drag_label = None
        self.__drag_start = (x, y)
        return True

    def on_mouse_release(self, x, y, buttons, modifiers):
        if not buttons & pyglet.window.mouse.LEFT:
            return False

        # we never thought we were dragging
        if not self.__dragging:
            return False

        # we might have been dragging, so clear it
        label, slot = self.__dragging
        self.__dragging = None

        # we didn't actually drag...
        if not self.drag_label:
            return False

        # we were dragging - clear the label
        self.remove(self.drag_label)
        self.drag_label = None

        # get the original item we're dragging
        item = label._item
        dx = self.__drag_start[0]

        body = self.character.body

        w, h = director.get_window_size()

        # if the item came from the body then remove it from there
        if dx < w//3:
            # NOOP drag to same area
            if x < w//3: return True

            body.equip(None, slot)

            # update source label
            label._item = None
            slabel = self.character.body.slot_dict[label._body_slot]
            label.element.text = slabel + ':'
        elif w//3 < dx < 2*w//3:
            # NOOP drag to same area
            if w//3 < x < 2*w//3: return True
            self.remove_from_pack(item)
        elif not self.party_items:
            raise ValueError("don't know where item came from")
        else:
            # NOOP drag to same area
            if x > 2*w//3: return True
            self.remove_from_party(item)

        if x < w//3:
            # user has dropped the item into a body slot
            for label in self.character_slots:
                if label._body_slot == item.slot:
                    break

            # move an existing item in that slot to the pack
            if label._item:
                body.equip(None, label._body_slot)
                self.add_to_pack(label._item)

            # update dest label
            label._item = item
            slabel = self.character.body.slot_dict[item.slot]
            label.element.text = '%s: %s'%(slabel, item.label)
            body.equip(item)

        elif x > 2*w//3 and self.parent.party_mode:
            self.add_to_party(item)
        elif w//3 < x < 2*w//3:
            self.add_to_pack(item)

        return True

    def add_to_pack(self, item):
        if len(self.character.carried) > 10: # XXX ARBITRARY NUMBER ALERT
            return self.add_to_party(item)
        elif self.pack_items:
            y = self.pack_items[-1].y - self.pack_items[-1].element.content_height
        else:
            y = self.pack_top
        w, h = director.get_window_size()
        l, y = self.add_label(item.label, w//3, y)
        l._item = item
        self.pack_items.append(l)
        self.pack_top = y

        self.character.carried.append(item)

    def remove_from_pack(self, item):
        self.character.carried.remove(item)
        h = 0
        for l in list(self.pack_items):
            if l._item == item:
                # remember height and remove it
                h = l.element.content_height
                self.remove(l)
                self.pack_items.remove(l)
                self.pack_top = l.y
            elif h:
                # shuffle the rest up
                l.y += h
                self.pack_top = l.y - l.element.content_height

    def add_to_party(self, item):
        for label in self.party_items:
            if item == label._item:
                # found existing item
                label._count += 1
                label.element.text = '%d x %s'%(label._count, item.label)
                return
        # add new item
        if self.party_items:
            y = self.party_items[-1].y - self.party_items[-1].element.content_height
        else:
            y = self.party_top
        w, h = director.get_window_size()
        l, y = self.add_label('1 x %s'%item.label, 2*w//3, y)
        l._item = item
        l._count = 1
        self.party_items.append(l)
        self.party_top = y

        self.party.items.append(item)

    def remove_from_party(self, item):
        self.party.items.remove(item)
        h = 0
        for l in list(self.party_items):
            self.party_top = l.y - l.element.content_height
            if l._item == item:
                l._count -= 1
                if l._count > 0:
                    l.element.text = '%d x %s'%(l._count, item.label)
                else:
                    # remember height and remove it
                    h = l.element.content_height
                    self.remove(l)
                    self.party_items.remove(l)
                    self.party_top = l.y
            elif h:
                # shuffle the rest up
                l.y += h

    def on_exit(self):
        for l in self.character_labels + self.character_slots + self.pack_items:
            self.remove(l)
        for l, skill in self.skills_labels:
            self.remove(l)
        self.remove(self.cur_xp)
        self.party_items = self.pack_items = None
        self.character_labels = self.character_slots = None
        self.character_pack = self.character_names = None

        self.music_player.pause()
        self.music_player = None
        self.shopping_music = None

        super(InventoryLayer, self).on_exit()



class InventoryScreen(cocos.scene.Scene):
    def __init__(self, party, character=None):
        super(InventoryScreen, self).__init__()
        self.party = party
        self.character = character

    def on_enter(self):
        super(InventoryScreen, self).on_enter()
        self.l = InventoryLayer(self.party, self.character)
        self.add(self.l)

    def on_exit(self):
        super(InventoryScreen, self).on_exit()
        self.remove(self.l)

class PartyInventoryScreen(InventoryScreen):
    party_mode = True

class CharacterInventoryScreen(InventoryScreen):
    party_mode = False



class BuySkillMenu(Menu):
    def __init__(self, title, character):
        super(BuySkillMenu, self).__init__(title)
        self.character = character

    def on_enter(self):
        super(BuySkillMenu, self).on_enter()
        self.menu_valign = CENTER
        self.menu_halign = CENTER

        self.font_item['color'] = (0, 0, 0, 255)
        self.font_item['font_size'] = 24
        self.font_item_selected['color'] = (150, 150, 0, 255)
        self.font_item_selected['font_size'] = 24
        self.font_title['color'] = (0, 0, 0, 255)

        # ugh, menu should DO this
        for child in self.get_children():
            self.remove(child)

        options = []
        for skill, info in character.skills.items():
            if skill in self.character.skills: continue
            options.append(MenuItem(info['label'], lambda skill=skill: self.add_skill(skill)))
        options.append(MenuItem('Cancel', self.on_quit))

        self.create_menu(options)

    def on_quit(self):
        director.pop()

    def add_skill(self, skill_name):
        self.character.skills[skill_name] = 1
        self.character.experience -= 5
        director.pop()

