import sys
import os
import random

import pyglet
import cocos
from cocos import tiles
from cocos.director import director
from cocos.actions import *
from cocos.menu import *
from cocos.tiles import RectCell, Resource

from gamelib import npc, ui, party, los, success, DummyPlayer

class CharacterLayer(tiles.ScrollableLayer):
    '''This is the layer that contains all the characters - player or otherwise.

    It manages the sprite for each character.
    '''
    def __init__(self):
        super(CharacterLayer, self).__init__()
        self.character_sprites = {}
        self.npc_sprites = {}

    def generate_sprite(self, character):
        sprite_name = character.sprite_name
        if character.is_dead():
            sprite_name = sprite_name.replace('.png', '-dead.png')
        image = pyglet.resource.image(sprite_name)
        image.anchor_x = image.width//2
        image.anchor_y = image.height//2

        map = self.parent.parent.ground
        s = cocos.sprite.Sprite(image, map.get_cell(*character.position).center)
        self.add(s, z=1)

        # add equipment in hands
        for item, offset in [(character.body.right_hand, 16),
                (character.body.left_hand, -16)]:
            if not item: continue
            i = pyglet.resource.image(item.sprite_name)
            i.anchor_x = i.width//2
            i.anchor_y = i.height//2
            item = cocos.sprite.Sprite(i, (offset, -8))
            s.add(item, z=1)

        return s

    def add_character(self, character):
        '''Add a player character to the character display layer.
        '''
        s = self.generate_sprite(character)
        self.character_sprites[character.name] = s

        # have the character let us know when things about it change
        character.push_handlers(self)

    def add_npc(self, npc):
        '''Add a non-player character to the character display layer.
        '''
        s = self.generate_sprite(npc)
        self.npc_sprites[npc.name] = s

        # have the NPC let us know when things about it change
        npc.push_handlers(self)

    def get_at_pixel(self, x, y, only_alive=True):
        '''Determine which character is at the given pixel. Must be a direct hit
        on the actual sprite of the character, not the whole cell the character is
        in.
        '''
        for list, mapping in [(self.parent.parent.npcs, self.npc_sprites),
                (self.parent.parent.party.characters, self.character_sprites)]:
            for thing in list:
                if not mapping[thing.name].contains(x, y): continue
                if only_alive and thing.is_dead(): continue
                return thing
        return None

    def get_at_cell(self, i, j, only_alive=True):
        '''Sometimes we need to ignore some characters.
        '''
        for list in (self.parent.parent.npcs, self.parent.parent.party.characters):
            for thing in list:
                if thing.position != (i, j): continue
                if only_alive and thing.is_dead(): continue
                return thing
        return None

    def on_exit(self):
        # clean up when this layer is no longer displayed
        super(CharacterLayer, self).on_exit()
        # XXX <abzde> erm, why? python garbage collects.
        #self.character_sprites = {}
        #self.npc_sprites = {}

    def get_character_sprite(self, character):
        '''From a Character instance, find the PC's Sprite.
        '''
        return self.character_sprites[character.name]
    def get_npc_sprite(self, character):
        '''From a Character instance, find the NPC's Sprite.
        '''
        return self.npc_sprites[character.name]
    def get_sprite(self, character):
        '''From a Character instance, find the PC or NPC Sprite.
        '''
        if isinstance(character, npc.NPC):
            return self.npc_sprites[character.name]
        else:
            return self.character_sprites[character.name]

    def on_character_move(self, character, old_position, position):
        '''A character (NPC or PC) has moved and is letting us know.
        '''
        # callback while we're not still attached to a Scene :(
        if not self.parent:
            return
        level = self.parent.parent
        cell = level.get_cell(*position)

        self.get_sprite(character).do(MoveTo(cell.center, .1))

        level.check_npc_visibility()

        if character is level.active_character:
            level.scroller.set_focus(*cell.center)

    def on_character_damage(self, character, old_health, damage):
        '''A character (NPC or PC) has taken damage.
        '''
        self.get_sprite(character).do(Blink(5, .5))

    def on_character_death(self, character):
        '''A character (NPC or PC) has taken damage.
        '''
        # use the same sprite name but with -dead appended before the suffix
        sprite_name = character.sprite_name.replace('.png', '-dead.png')
        image = pyglet.resource.image(sprite_name)
        image.anchor_x = image.width//2
        image.anchor_y = image.height//2

        s = self.get_sprite(character)
        s.image = image

        # need the sprite to drop to the back, and this is the only way
        self.remove(s)
        self.add(s, z=0)

        # check success
        level = self.parent.parent
        if not [n for n in level.npcs if not n.is_dead()]:
            director.replace(success.Success(level))
            return

# XXX this belongs in the tactical ui but the recent
# XXX refactoring makes that harder
class GroundInfoLayer(tiles.ScrollableLayer):
    shown = None
    def update_footsteps(self, positions):
        if self.shown is None:
            self.shown = {}

        image = pyglet.resource.image('icon-walk.png')

        # remove old
        sp = set(positions)
        for pos in dict(self.shown):
            if pos not in sp:
                self.remove(self.shown[pos])
                del self.shown[pos]

        # add new
        m = self.parent.parent.ground
        for pos in positions:
            if pos in self.shown:
                continue
            cell = m.get_cell(*pos)
            s = cocos.sprite.Sprite(image, cell.center)
            self.add(s)
            self.shown[pos] = s

    def on_exit(self):
        super(GroundInfoLayer, self).on_exit()
        self.shown = None


class Level(cocos.scene.Scene):
    '''

    How the tactical game works
    ---------------------------

    The player and computer take turns.

    The player's turn consists of:

    1. invoke start_new_turn on all of the player's characters
    2. open up the tactical_ui Interface for the first character
    3. handle any actions invoked through the Interface
    4. when space is hit, load up that Interface for the next character
    5. once all characters have been handled, the game switches to NPC turn

    The NPC turn consists of:

    1. invoke start_new_turn on all of the player's characters
    2. invoke figure_what_to_do for each NPC in turn*
    3. once all NPCs are handled switch back to PC turn

    * important to note here is that the NPCs aren't just handled in a loop
    as that might consume the CPU for some noticeable period of time. Instead
    after each NPC is handled we set up a callback to handled another one.
    This gives the rest of the program (rendering, event handling) a chance
    to do something in between handling NPCs.

    '''

    PLAYER_TURN = 'player'
    NPC_TURN = 'npc'

    def __init__(self, player_party, name):
        super(Level, self).__init__()

        self.party = player_party

        # Basic scrolling manager. All layers added to this layer will be scrolled
        # when we set_focus() on it.
        self.scroller = tiles.ScrollingManager()
        self.add(self.scroller)

        # add in the level's ground layer loaded by the tile code
        name = name.replace('\\', '/')
        path = pyglet.resource.location(name).path
        fname = path + '/' + name
        self.ground = tiles.load(fname)['ground']
        self.scroller.add(self.ground)

        # add in the info layer for display over the ground
        self.ground_info = GroundInfoLayer()
        self.scroller.add(self.ground_info, z=self.ground.origin_z+1)

        # add in NPCs and figure the PC spawn points using the properties
        # from the XML map file
        self.npcs = []
        self.spawns = []

        # XXX the following is for creating the map when first encountered, we also
        # XXX need a load/save
        for column in self.ground.cells:
            for cell in column:
                if cell.get('npc-class'):
                    self.npcs.append(npc.create(cell))
                elif cell.get('player-spawn'):
                    self.spawns.append(cell)

        # start PCs in random spots
        random.shuffle(self.spawns)

        # a layer to display character stuff (their sprites, mostly)
        self.character_layer = CharacterLayer()
        self.scroller.add(self.character_layer, z=self.ground.origin_z+2)
        for n in self.npcs:
            self.character_layer.add_npc(n)
        self.active_character = None

    def get_at_cell(self, i, j, ignore=None):
        c = self.character_layer.get_at_cell(i, j)
        return c

    def get_cell(self, i, j, ignore=None):
        '''Sometimes (ok, just in map solving) we need to ignore some contents of the
        level.
        '''
        ground_cell = self.ground.get_cell(i, j)

        # see whether a character affects that cell
        c = self.character_layer.get_at_cell(i, j)
        if ignore and c in ignore:
            cell = ground_cell
        elif c:
            # make this cell impassable
            d = ground_cell.properties.copy()
            d['impassable'] = True
            cell = RectCell(i, j, 64, 64, d, ground_cell.tile)
        else:
            cell = ground_cell
        return cell

    UP = (0, 1)
    DOWN = (0, -1)
    LEFT = (-1, 0)
    RIGHT = (1, 0)
    def get_neighbor(self, cell, direction, ignore=None):
        '''Sometimes (ok, just in map solving) we need to ignore some contents of the
        level.
        '''
        dx, dy = direction
        return self.get_cell(cell.i + dx, cell.j + dy, ignore=ignore)

class CrawlLevel(Level):
    def __init__(self, player_party, name):
        super(CrawlLevel, self).__init__(player_party, name)
        self.party.mode = 'crawl'
        self.spawn = self.spawns.pop()
        self.active_character = self.party.characters[0]
        self.active_character.position = (self.spawn.i, self.spawn.j)
        self.character_layer.add_character(self.active_character)

    def on_enter(self):
        super(CrawlLevel, self).on_enter()

        self.player_ui = ui.CrawlInterface(self.active_character)
        self.add(self.player_ui, z=10)
        self.active_character.push_handlers(self.player_ui)
        self.active_character.push_handlers(self)

        self.start_player_turn()

    def on_exit(self):
        self.active_character.pop_handlers()
        self.active_character.pop_handlers()

        self.remove(self.player_ui)
        self.player_ui = None

    def start_player_turn(self):
        self.mode = self.PLAYER_TURN


        # and re-focus the display on the center of the character's cell
        cell = self.ground.get_cell(*self.active_character.position)
        self.scroller.set_focus(*cell.center)

    def next_player_character(self):
        self.active_character.pop_handlers()
        self.active_character.pop_handlers()
        self.start_player_turn()

class TacticalLevel(Level):
    def __init__(self, *a, **kw):
        super(TacticalLevel, self).__init__(*a, **kw)
        self.party.mode = 'tactical'
        for character in self.party.characters:
            s = self.spawns.pop()
            character.position = (s.i, s.j)
            self.character_layer.add_character(character)
        self.player_ui = None

    def on_enter(self):
        super(TacticalLevel, self).on_enter()

        # add the user interface for interacting with the character
        # not added to scrolling manager so it's not scrolled
        self.player_ui = ui.TacticalInterface(None)
        self.add(self.player_ui, z=10)

        # make the tactical UI listen for player events (like actions
        # used, or DEATH :)
        for character in self.party.characters:
            character.push_handlers(self.player_ui)
            character.push_handlers(self)

        self.check_npc_visibility()

        self.start_player_turn()

        try:
            self.music_player = pyglet.media.Player()
            self.menu_music = pyglet.resource.media('music_dungeon_t.ogg')
        except pyglet.media.riff.WAVEFormatException:
            self.music_player = DummyPlayer()
            self.menu_music = None

        self.music_player.queue(self.menu_music)
        self.music_player.eos_action = self.music_player.EOS_LOOP
        self.music_player.play()

    def on_character_death(self, character):
        # check failure
        if not [c for c in self.party.characters if not c.is_dead()]:
            director.replace(cocos.scene.Scene(Failure('You Failed!')))
            return

    def on_exit(self):
        super(TacticalLevel, self).on_exit()

        for character in self.party.characters:
            # pop me
            character.pop_handlers()
            # pop the player ui
            character.pop_handlers()

        self.remove(self.player_ui)
        self.player_ui = None

        # XXX fade volume down and fade menu out
        self.music_player.pause()
        self.music_player = None
        self.menu_music = None

    def start_player_turn(self):
        '''Called when the player takes control of the game.
        '''
        self.mode = self.PLAYER_TURN

        # initialise their per-turn state
        for c in self.party.characters:
            c.start_new_turn()
            self.player_ui.update_char_consumables(c)

        # queue up the first character for control
        self.active_char_index = -1
        self.next_player_character()

    def next_player_character(self):
        '''Invoked to line up a new character for the player to control, or to pass
        control to NPC when the player is all done.
        '''

        # get the next non-dead character in sequence
        if not [c for c in self.party.characters if not c.is_dead()]:
            print 'ALL DEAD! GAME OVER!'
        while 1:
            self.active_char_index = (self.active_char_index + 1)%4
            c = self.party.characters[self.active_char_index]
            if not c.is_dead():
                break

        # pop new character from the start of the list
        self.active_character = c
        self.player_ui.activate_character(c)

        # and re-focus the display on the center of the character's cell
        cell = self.ground.get_cell(*c.position)
        self.scroller.set_focus(*cell.center)

    def end_turn(self):
        '''Invoked to end the player's turn and pass control over to NPC.
        '''
        self.player_ui.activate_character(None)

        # the player is all done, switch to NPC
        self.mode = self.NPC_TURN

        # figure NPCs to process
        npc_left = [c for c in self.npcs if c.is_active()]
        self.npc_to_process = npc_left
        for n in npc_left:
            n.start_new_turn()

        # Add a callback to process the next NPC so that the rest of
        # the game (user event handling, rendering, animations, etc)
        # can have a go - Cocos will automatically call this as part
        # of the next clock tick in its event loop.
        self.do(Delay(0.01) + CallFunc(self.next_non_player_character))

    def next_non_player_character(self):
        '''Invoked to process the turn for a non-player-character, or to pass control
        back to the player if the NPCs are all done.
        '''
        if self.npc_to_process:
            n = self.npc_to_process.pop()
            s = self.character_layer.get_npc_sprite(n)
            n.figure_what_to_do(self)
            self.do(Delay(0.01) + CallFunc(self.next_non_player_character))
        else:
            self.start_player_turn()

    def check_npc_visibility(self):
        for npc in self.npcs:
            visible = []
            for c in self.party.characters:
                src  = c.position
                dest = npc.position
                visible.append(bool(los.line_of_sight(self, src, dest, True)))

            if any(visible):
                o = 255
            elif '--god' in sys.argv:
                o = 100
            else:
                o = 0

            s = self.character_layer.get_npc_sprite(npc)
            s.opacity = o
            for s in s.get_children():
                if isinstance(s, cocos.sprite.Sprite):
                    s.opacity = o


class Failure(Menu):
    def on_enter(self):
        super(Failure, self).on_enter()
        self.menu_valign = CENTER
        self.menu_halign = CENTER

        self.font_item['color'] = (0, 0, 0, 255)
        self.font_item_selected['color'] = (150, 150, 0, 255)
        self.font_title['color'] = (0, 0, 0, 255)

        # ugh, menu should DO this
        for child in self.get_children():
            self.remove(child)

        options = [
            MenuItem('Press enter', self.on_quit),
        ]

        self.create_menu(options)

    def on_quit(self):
        director.pop();director.pop()

