import random
import math
from xml.etree import ElementTree

import pyglet

import cocos
from cocos.actions import *

from gamelib import actions

class Spell(object):
    # Spells have a number of attributes
    # name          -- the spell's name (also its label)
    # type          -- the school of magic
    # description   -- a longer description of its effect
    # mana_cost     -- the mana cost
    # action_cost   -- the action points cost
    # difficulty    -- the relative difficulty (between 1 for easy and 0 for hardest)
    # action_class  -- the Action class to use
    # requires_target -- whether a ranged spell requires a character target

    def do(self, level, character, cell, other):
        '''Make this spell do something as cast by character on cell or other
        in level.

        Default implementation is to use the mana points.
        '''
        return character.use_mana_points(self.cost)

    def check(self, action):
        '''The Action class wishes to make sure this spell is OK with
        being cast. The action has a character, cell and other (character)
        attribute. If the spell is ranged there is also a distance
        attribute.
        '''
        return True

    def animate_attack(self, layer, attack):
        layer.do(CallFunc(layer.done))

    def to_xml(self, root):
        m = ElementTree.SubElement(root, 'spell', type=self.type,
            name=self.name)
        m.tail = '\n'

def from_xml(tag):
    '''Load the item from an XML file.
    '''
    type = tag.get('type')
    name = tag.get('name')
    return spells[type][name]()

spells = {}
def add_spell(klass):
    spells.setdefault(klass.type, {})[klass.name] = klass

class MagicMissile(Spell):
    name = 'Magic Missile'
    type = 'attack'
    description = 'A bolt of magical energy to strike your foe'
    mana_cost = 1
    action_cost = 3
    action_class = actions.RangedMagic
    difficulty = 1                  # always succeeds
    requires_target = True

    def do(self, level, character, cell, other):
        # XXX plus skill
        damage = random.randint(1, 7)
        print 'MagicMissile:', damage, 'damage to', other
        other.apply_damage(damage)
        return True

    def animate_attack(self, layer, attack):
        i1, j1 = attack.character.position
        x1, y1 = attack.level.ground.get_cell(i1, j1).center

        i2, j2 = attack.other.position
        x2, y2 = attack.level.ground.get_cell(i2, j2).center

        angle = math.degrees(math.atan2(y1-y2, x2-x1))

        i = pyglet.resource.image('effect-magic-missile.png')
        s = cocos.sprite.Sprite(i, (x1, y1), rotation=angle)
        layer.add(s)
        # XXX need constant-speed not constant-time animation here
        s.do(MoveTo((x2, y2), .5) + CallFunc(layer.done))
add_spell(MagicMissile)

class BurningHands(Spell):
    name = 'Burning Hands'
    type = 'attack'
    description = 'Your hands burn with magical fury!'
    mana_cost = 1
    action_cost = 3
    action_class = actions.MeleeMagic
    difficulty = 1                  # always succeeds

    def do(self, level, character, cell, other):
        # XXX plus skill
        damage = random.randint(2, 9)
        print 'BurningHands:', damage, 'damage to', other
        other.apply_damage(damage)
        return True

    def animate_attack(self, layer, attack):
        i1, j1 = attack.character.position
        x1, y1 = attack.level.ground.get_cell(i1, j1).center

        i2, j2 = attack.other.position
        x2, y2 = attack.level.ground.get_cell(i2, j2).center

        angle = math.degrees(math.atan2(y1-y2, x2-x1))

        pos = (x1 + (x2-x1)/2, y1 + (y2-y1)/2)

        i = pyglet.resource.image('effect-burning-hands.png')
        i.anchor_x = 0
        s = cocos.sprite.Sprite(i, pos, rotation=angle)
        layer.add(s)

        fade = Accelerate(FadeIn(.25))
        s.do(fade + Reverse(fade) + CallFunc(layer.done))
add_spell(BurningHands)

class Kill(Spell):
    name = 'Kill'
    type = 'attack'
    description = 'You kill your enemy through sheer force of will'
    mana_cost = 100
    action_cost = 5
    action_class = actions.RangedMagic
    difficulty = 0
    requires_target = True

    def do(self, level, character, cell, other):
        print 'Kill:', other
        other.apply_damage(other.health)
        return True
add_spell(Kill)


class Heal(Spell):
    name = 'Heal'
    type = 'healing'
    description = 'You heal your target'
    mana_cost = 1
    action_cost = 2
    action_class = actions.MeleeMagic
    difficulty = 1
    requires_target = True

    def do(self, level, character, cell, other):
        # XXX plus skill
        healing = random.randint(2, 9)
        print 'Heal:', other, healing
        other.apply_damage(-healing)
        return True


    def animate_attack(self, layer, attack):
        i, j = attack.other.position
        x, y = attack.level.ground.get_cell(i, j).center

        fade = (Accelerate(FadeIn(.25)) + FadeOut(.25)) | ScaleTo(2, .5)

        i = pyglet.resource.image('effect-heal.png')
        for n, (dx, dy) in enumerate([(-16, -16), (16, 24), (24, -40), (8, 8)]):
            pos = (x + dx, y + dy)
            s = cocos.sprite.Sprite(i, pos, opacity=0)
            layer.add(s)
            anim = Delay(n * .25) + fade
            if n == 3:
                anim = anim + CallFunc(layer.done)
            s.do(anim)
add_spell(Heal)

