import pyglet

import cocos
from cocos.director import director
from cocos.actions import *
from gamelib import items, DummyPlayer
from random import randint

class ShoppingLayer(cocos.layer.Layer):
    is_event_handler = True

    def __init__(self, party):
        super(ShoppingLayer, self).__init__()
        self.party = party
        self.shop_items = items.random_shop(party)

    def add_heading(self, text, x, y):
        return self.add_label(text, x, y, True)

    def add_label(self, text, x, y, bold=False):
        l = cocos.text.Label(text, (x, y), bold=bold, color=(0, 0, 0, 255),
            anchor_y='top')
        self.add(l)
        return l, y - l.element.content_height

    def on_enter(self):
        super(ShoppingLayer, self).on_enter()

        self.batch = pyglet.graphics.Batch()

        w, h = director.get_window_size()

        x = 0
        l, y = self.add_heading('Shop Items', x, h)

        try:
            self.music_player = pyglet.media.Player()
            self.shopping_music = pyglet.resource.media('music_town_t.ogg')
        except pyglet.media.riff.WAVEFormatException:
            self.music_player = DummyPlayer()
            self.shopping_music = None

        self.music_player.queue(self.shopping_music)
        self.music_player.eos_action = self.music_player.EOS_LOOP
        self.music_player.play()

        # display the items, sorting into duplicates
        l = items.count_duplicates(self.shop_items)
        self.shop_labels = []
        for item, count in l:
            l, y = self.add_label('%d x %s - %d gold'%(count, item.label, item.cost), x, y)
            l._item = item
            l._count = count
            l._cost = item.cost
            self.shop_labels.append(l)

        self.shop_top = y

        # party items appear in the second column
        x = w//2
        l, y = self.add_heading('Party Items', x, h)
        l, y = self.add_label('%d gold coins'%(self.party.coins), x, y)
        self.party_coins = l
        # display the items, sorting into duplicates
        l = items.count_duplicates(self.party.items)
        self.party_labels = []
        for item, count in l:
            l, y = self.add_label('%d x %s - %d gold'%(count, item.label, item.cost), x, y)
            l._item = item
            l._count = count
            l._cost = item.cost
            self.party_labels.append(l)
        self.party_top = y

        l, y = self.add_heading('Hints:', w//4, h//4)
        l, y = self.add_label('Drag and drop items to move them around (buy/sell).', w//4, y)
        l, y = self.add_label('Press <esc> when you are finished.', w//4, y)


    def on_exit(self):
        self.party_labels = None
        self.shop_labels = None
        super(ShoppingLayer, self).on_exit()

        self.music_player.pause()
        self.music_player = None
        self.shopping_music = None

    def find_hit(self, x, y):
        for label in self.shop_labels:
            if x < label.x: continue
            if y > label.y: continue
            if x > label.x + label.element.content_width: continue
            if y < label.y - label.element.content_height: continue
            return label

        for label in self.party_labels:
            if x < label.x: continue
            if y > label.y: continue
            if x > label.x + label.element.content_width: continue
            if y < label.y - label.element.content_height: continue
            return label
        return None


    __over = None
    def on_mouse_motion(self, x, y, dx, dy):
        hit = self.find_hit(x, y)
        if not hit:
            if self.__over is not None:
                self.__over.element.color = (0, 0, 0, 255)
                self.__over = None
            return True

        if hit is not self.__over:
            if self.__over is not None:
                self.__over.element.color = (0, 0, 0, 255)
            self.__over = hit
            hit.element.color = (100, 100, 0, 255)

        return True

    __dragging = None
    __drag_start = None
    drag_label = None
    def on_mouse_drag(self, x, y, dx, dy, buttons, modifiers):
        if not self.__dragging:
            # do the mouse-over highlighting
            self.on_mouse_motion(x, y, dx, dy)
            return True

        # XXX highlight the section of screen active

        if self.drag_label:
            self.drag_label.position = (x, y)
            return True

        dx, dy = self.__drag_start
        if abs(x - dx) + abs(y - dy) < 2:
            return True

        item = self.__dragging
        self.drag_label = cocos.text.Label(item._item.label, (x, y),
            color=(0, 0, 0, 255))
        self.add(self.drag_label)
        return True

    def on_mouse_press(self, x, y, buttons, modifiers):
        if not buttons & pyglet.window.mouse.LEFT:
            return False

        hit = self.find_hit(x, y)
        if not hit:
            return False

        self.__dragging = hit
        self.drag_label = None
        self.__drag_start = (x, y)
        return True

    def on_mouse_release(self, x, y, buttons, modifiers):
        if not buttons & pyglet.window.mouse.LEFT:
            return False

        # we never thought we were dragging
        if not self.__dragging:
            return False

        # we might have been dragging, so clear it
        label = self.__dragging
        self.__dragging = None

        # we didn't actually drag...
        if not self.drag_label:
            return False

        # we were dragging - clear the label
        self.remove(self.drag_label)
        self.drag_label = None

        # get the original item we're dragging
        item = label._item
        dx = self.__drag_start[0]

        w, h = director.get_window_size()

        if dx < w//2:
            # NOOP drag to same area
            if x < w//2: return True
            self.remove_from_shop(item)
        else:
            # NOOP drag to same area
            if x > w//2: return True
            self.remove_from_party(item)

        if x < w//2:
            self.add_to_shop(item)
        else:
            self.add_to_party(item)
        return True

    def add_to_shop(self, item):
        for label in self.shop_labels:
            if item == label._item:
                # found existing item
                label._count += 1
                label.element.text = '%d x %s - %d gold'%(label._count, item.label, item.cost)
                self.party.coins += item.cost
                self.party_coins.element.text = '%d gold coins'%(self.party.coins)
                return

        # add new item
        l, y = self.add_label('1 x %s - %d gold'%(item.label, item.cost),  0, self.shop_top)
        l._item = item
        l._count = 1
        self.shop_labels.append(l)
        self.shop_top = y - l.element.content_height
        self.party.coins += item.cost
        self.party_coins.element.text = '%d gold coins'%(self.party.coins)
        self.shop_items.append(item)

    def remove_from_shop(self, item):
        if self.party.coins >= item.cost:
            self.shop_items.remove(item)
            h = 0
            for l in list(self.shop_labels):
                if l._item == item:
                    l._count -= 1
                    if l._count > 0:
                        l.element.text = '%d x %s - %d gold'%(l._count, item.label, item.cost)
                    else:
                        # remember height and remove it
                        h = l.element.content_height
                        self.remove(l)
                        self.shop_labels.remove(l)
                        self.shop_top = l.y
                elif h:
                    # shuffle the rest up
                    l.y += h
                    self.shop_top = l.y - l.element.content_height


    def add_to_party(self, item):
        if self.party.coins >= item.cost:
            self.party.coins -= item.cost
            self.party_coins.element.text = '%d gold coins'%(self.party.coins)
            for label in self.party_labels:
                if item == label._item:
                    # found existing item
                    label._count += 1
                    label.element.text = '%d x %s - %d gold'%(label._count, item.label, item.cost)
                    return

            # add new item
            w, h = director.get_window_size()
            l, y = self.add_label('1 x %s - %d cost'%(item.label, item.cost),  w//2, self.party_top)
            l._item = item
            l._count = 1
            self.party_labels.append(l)
            self.party_top = y - l.element.content_height

            self.party.items.append(item)
            
        else:
            red = CallFunc(lambda: self.party_coins.element._set_color((255, 0, 0, 255)))
            blank = CallFunc(lambda: self.party_coins.element._set_color((255, 255, 255, 255)))
            black = CallFunc(lambda: self.party_coins.element._set_color((0, 0, 0, 255)))
            self.do((red + Delay(0.2) + blank + Delay(0.2))*5 + black)
            
    def remove_from_party(self, item):
        self.party.items.remove(item)
        h = 0
        for l in list(self.party_labels):
            if l._item == item:
                l._count -= 1
                if l._count > 0:
                    l.element.text = '%d x %s - %d gold'%(l._count, item.label, l._cost)
                else:
                    # remember height and remove it
                    h = l.element.content_height
                    self.remove(l)
                    self.party_labels.remove(l)
                    self.party_top = l.y
            elif h:
                # shuffle the rest up
                l.y += h
                self.party_top = l.y - l.element.content_height

