'''This module contains the interface for the "tactical" game, specifically for
controlling one of the player characters in a dungeon setting.
'''

import pyglet
import cocos
from cocos.director import director
from cocos.actions import *

from gamelib import popup_menu, inventory

class Interface(cocos.layer.Layer):
    '''This interface is *always* a direct child of the Level Scene (as defined
    in the level module.

    This means the "parent" attribute is that Level, giving us access to the
    map information and all the other characters etc.
    '''
    is_event_handler = True

    def __init__(self, character):
        super(Interface, self).__init__()
        self.character = character

    def on_text_motion(self, motion):
        map = self.parent
        i, j = self.character.position
        cell = map.get_cell(i, j, ignore=[self.character])
        new_cell = None
        if motion == pyglet.window.key.MOTION_LEFT:
            new_cell = map.get_neighbor(cell, map.LEFT)
            if cell.get('blocks-left') or new_cell.get('impassable'):
                new_cell = None
        elif motion == pyglet.window.key.MOTION_RIGHT:
            new_cell = map.get_neighbor(cell, map.RIGHT)
            if cell.get('blocks-right') or new_cell.get('impassable'):
                new_cell = None
        elif motion == pyglet.window.key.MOTION_DOWN:
            new_cell = map.get_neighbor(cell, map.DOWN)
            if cell.get('blocks-down') or new_cell.get('impassable'):
                new_cell = None
        elif motion == pyglet.window.key.MOTION_UP:
            new_cell = map.get_neighbor(cell, map.UP)
            if cell.get('blocks-up') or new_cell.get('impassable'):
                new_cell = None

        if new_cell is not None:
            # XXX figure actual cost here when we have real costs
            self.character.move_to((new_cell.i, new_cell.j), 1)

        return True

class CrawlInterface(Interface):
    def on_text(self, text):
        super(CrawlInterface, self).on_text(text)
        if text == 'i':
            director.push(inventory.PartyInventoryScreen(self.parent.party))
            return True


class TacticalInterface(Interface):
    def __init__(self, character):
        super(TacticalInterface, self).__init__(character)

    def on_enter(self):
        super(TacticalInterface, self).on_enter()

        self.batch = pyglet.graphics.Batch()

        w, h = director.get_window_size()

        self.character_info = {}
        ly = h
        for character in self.parent.party.characters:
            self.character_info[character.name] = d = {}
            y = ly

            if character is self.character:
                color = (255, 255, 255, 255)
            else:
                color = (255, 255, 255, 128)
            d['bg'] = self.batch.add(4, pyglet.gl.GL_QUADS, None,
                ('v2i', (0, y, 0, y-80, 100, y-80, 100, y)),
                ('c4B', color * 4)
            )

            y -= 4

            image = pyglet.resource.image(character.sprite_name)
            scale = 32. / image.height
            d['sprite'] = cocos.sprite.Sprite(image, (20, y-16), scale=scale)
            self.add(d['sprite'], z=1)

            d['ap'] = pyglet.text.Label('AP: %d/%d'%(character.actions_remaining,
                character.max_actions), font_size=10, x=36, y=y, anchor_y='top',
                color=(0, 0, 0, 255), batch=self.batch)
            y -= d['ap'].content_height
            d['hp'] = pyglet.text.Label('HP: %d/%d'%(character.health,
                character.max_health), font_size=10, x=36, y=y, anchor_y='top',
                color=(0, 0, 0, 255), batch=self.batch)
            y -= d['hp'].content_height
            d['mp'] = pyglet.text.Label('MP: %d/%d'%(character.mana,
                character.max_mana), font_size=10, x=36, y=y, anchor_y='top',
                color=(0, 0, 0, 255), batch=self.batch)
            y -= d['mp'].content_height

            d['name'] = pyglet.text.Label(character.name, x=4, y=max(34, y),
                anchor_y='top', color=(0, 0, 0, 255), batch=self.batch)

            ly -= 84

        y = ly - 76

        # buttons now
        image = pyglet.resource.image('button-manage-party.png')
        self.party_button = cocos.sprite.Sprite(image, (2, y), anchor=(0, 0))
        self.add(self.party_button)
        y -= (image.height + 8)

        image = pyglet.resource.image('button-next-char.png')
        self.next_char_button = cocos.sprite.Sprite(image, (2, y), anchor=(0, 0))
        self.add(self.next_char_button)
        y -= (image.height + 8)

        image = pyglet.resource.image('button-end-turn.png')
        self.next_turn_button = cocos.sprite.Sprite(image, (2, y), anchor=(0, 0))
        self.add(self.next_turn_button)
        y -= (image.height + 8)

        image = pyglet.resource.image('button-retreat.png')
        self.retreat_button = cocos.sprite.Sprite(image, (2, y), anchor=(0, 0))
        self.add(self.retreat_button)
        y -= (image.height + 8)


    def activate_character(self, character):
        if self.character is not None:
            vl = self.character_info[self.character.name]['bg']
            vl.colors[:] = (255, 255, 255, 128) * 4

        self.character = character
        if character is not None:
            vl = self.character_info[self.character.name]['bg']
            vl.colors[:] = (255, 255, 255, 255) * 4
            self.update_movable_highlight()

    def on_exit(self):
        super(TacticalInterface, self).on_exit()
        for d in self.character_info.values():
            self.remove(d['sprite'])
        self.character_info = {}
        self.batch = None

    def on_text(self, text):
        if text == 'i':
            director.push(inventory.CharacterInventoryScreen(self.parent.party,
                self.character))
            return True
        if text == ' ':
            self.parent.next_player_character()
            return True

    def on_key_press(self, key, modifier):
        if key == pyglet.window.key.ENTER:
            self.parent.end_turn()
            return True

    def on_mouse_press(self, x, y, button, modifiers):
        if button & pyglet.window.mouse.RIGHT:
            # Right-click action menu

            # translate screen pixel to map pixel
            mx, my = self.parent.scroller.pixel_from_screen(x, y)

            # get the cell
            cell = self.parent.ground.get_at_pixel(mx, my)

            # and any creature in that cell
            # XXX one day we might have actions to perform on corpses
            other = self.parent.character_layer.get_at_pixel(mx, my, only_alive=True)

            # right, now get the action options for the character
            options = self.character.get_action_options()
            menu = []
            for o in options:
                # run the check first as it might update cost, label, who knows
                ok = o.check(self.parent, self.character, cell, other)

                label = '%s (%d AP)'%(o.label, o.cost)
                if not ok or o.chance == 0:
                    menu.append(((o.type, o.name), label, False))
                elif o.chance != 1:
                    label = '%s %d%% (%d AP)'%(o.label, o.chance * 100, o.cost)
                    menu.append(((o.type, o.name), label, True))
                else:
                    menu.append(((o.type, o.name), label, True))

            if not menu:
                return

            def on_close(selected, options=options):
                for o in options:
                    if (o.type, o.name) == selected:
                        o.do()
                        return

            menu = popup_menu.PopupMenu(x, y, menu, on_close)
            self.add(menu, z=1)
            return True

        if button & pyglet.window.mouse.LEFT:
            # Left clicking player info window to change character
            # figure out which char was clicked
            w, h = director.get_window_size()
            if x > 100: return
            if y > h - 84:
                self.activate_character(self.parent.party.characters[0])
            elif y > h - 168:
                self.activate_character(self.parent.party.characters[1])
            elif y > h - 252:
                self.activate_character(self.parent.party.characters[2])
            elif y > h - 336:
                self.activate_character(self.parent.party.characters[3])

            elif self.party_button.y < y < self.party_button.y + self.party_button.height:
                director.push(inventory.CharacterInventoryScreen(self.parent.party,
                    self.character))
            elif self.next_char_button.y < y < self.next_char_button.y + self.next_char_button.height:
                self.parent.next_player_character()
            elif self.next_turn_button.y < y < self.next_turn_button.y + self.next_turn_button.height:
                self.parent.end_turn()
            elif self.retreat_button.y < y < self.retreat_button.y + self.retreat_button.height:
                director.pop()
            else:
                return False

            return True

    def update_char_consumables(self, character):
        self.character_info[character.name]['ap'].text = 'AP: %d/%d'%(
            character.actions_remaining, character.max_actions)
        self.character_info[character.name]['mp'].text = 'MP: %d/%d'%(
            character.mana, character.max_mana)

    def on_character_action_used(self, character, points_left, points_used):
        self.character_info[character.name]['ap'].text = 'AP: %d/%d'%(
            character.actions_remaining, character.max_actions)
        self.update_movable_highlight()
        # no return in case someone else wants the event

    def on_character_mana_used(self, character, points_left, points_used):
        self.character_info[character.name]['mp'].text = 'MP: %d/%d'%(
            character.mana, character.max_mana)
        # no return in case someone else wants the event

    def on_character_damage(self, character, old_health, damage):
        self.character_info[character.name]['hp'].text = 'HP: %d/%d'%(
            character.health, character.max_health)
        # no return in case someone else wants the event

    def on_character_move(self, character, old_pos, new_pos):
        # highlight the cells the player may move to
        self.update_movable_highlight()

    def update_movable_highlight(self):
        # somehow invoked during init
        if not self.character:
            return
        level = self.parent
        l = [cell.position for cell in self.character.find_accessible_cells(level)]
        level.ground_info.update_footsteps(l)
        # no return in case someone else wants the event

    def draw(self):
        self.transform()
        self.batch.draw()

